class BlocksController < ApplicationController
  include MachineResponse

  def self.set_target_before
    super + [:reactivate, :discard, :wash, :force_to_clean]
  end

  model_data_store(Block)

  def before_set_targets
    Block.left_outer_joins(:job)
  end

  def before_filter(filter)
    if filter[:id] == 'job_number'
      filter[:id] = 'job.number'
    end
    filter
  end

  def before_sort(sort)
    if sort[:id] == 'job_number'
      sort[:id] = 'job.number'
    end
    sort
  end

  def add
    begin
      AddBlockJob.perform_later(add_params)
    rescue => e
      Rails.logger.error e
    end
    send_machine_response
  end

  def import
    blocks = params[:blocks]

    if blocks.nil?
      return render json: {error: "Invalid Block Data"}, status: :unprocessable_entity
    end

    settings = BlockReuseSetting.first

    begin
      Block.transaction do
        blocks.each do |block|
          dmc = block[:dmc]
          usage_count = block[:usage_count]

          unless dmc.nil? || usage_count.nil?
            block_record = Block.find_by(dmc: dmc)

            if block_record.nil?
              block_record = Block.new(dmc: dmc, usage_count: 0)
            end

            block_record.usage_count = block_record.usage_count + usage_count

            # Note: Do not use method 'block.use_wash_cycle' and 'block.use_discard_cycle' as that would cause 3 db transaction for each block
            discard_usage = block_record.new_record? ? settings.block_discard_after : block_record.cycles_until_discard
            block_discard_cycles = discard_usage - usage_count
            block_record.cycles_until_discard = block_discard_cycles > 0 ? block_discard_cycles : 0

            # Cycles_until_wash is cyclic, so we have to figure out how many are left using the usage count mod. the max usage per cycle
            wash_usage = block_record.new_record? ? settings.block_wash_after : block_record.cycles_until_wash
            block_record.cycles_until_wash = wash_usage - (usage_count % settings.block_wash_after)

            block_record.save!
          end
        end
      end
    rescue => err
      render json: {error: "Unable to import data: #{err}"}, status: :unprocessable_entity
    end
  end

  def reactivate
    settings = BlockReuseSetting.first

    return render json: {error: "error with block settings"}, status: :internal_server_error if settings.nil?

    @model.cycles_until_discard = settings.block_discard_after

    if @model.save
      BlockReactivateHistory.create(block: @model, action: "reactivate")
      render :show
    else
      render json: @model.errors, status: :unprocessable_entity
    end
  end

  def discard
    @model.cycles_until_discard = 0

    if @model.save
      BlockReactivateHistory.create(block: @model, action: "discard")
      render :show
    else
      render json: @model.errors, status: :unprocessable_entity
    end
  end

  def wash
    settings = BlockReuseSetting.first

    return render json: {error: "error with block settings"}, status: :internal_server_error if settings.nil?

    @model.cycles_until_wash = settings.block_wash_after

    if @model.save
      render :show
    else
      render json: @model.errors, status: :unprocessable_entity
    end
  end

  def force_to_clean
    @model.cycles_until_wash = 0

    if @model.save
      render :show
    else
      render json: @model.errors, status: :unprocessable_entity
    end
  end

  def find_by_dmc
    dmc = params[:dmc]

    return render json: {error: "No DMC provided."}, status: :not_found if dmc.nil?

    @model = Block.find_by(dmc: dmc)

    return render json: {error: "Block with matching DMC not found."}, status: :not_found if @model.nil?

    render :show
  end

  private

  def add_params
    params[:block] = params[:blockInfo] if params[:blockInfo]
    block = params.require(:block)
    side_num = Sideable.side_letter_to_side_num(block[:lensside])
    job = FindHelpers::JobFinders.find_active_or_create_by_number_and_side_num(block[:job], side_num)
    {
      dmc: block[:dmc],
      side: side_num,
      job: job,
      rfid_tag: block[:rfidtagid]
    }
  end
end
