class JobBreakagesController < ApplicationController
  class JobDoesNotExist < StandardError; end

  def self.set_target_before
    super + [:validate]
  end


  model_data_store(JobBreakage, actions = {create: true, read: true, update: false, delete: false, index: true},
                   permissions = {create: true, read: true, update: false, delete: false, index: true})
  also_protect_with([RxModule::Authorization.verb("validate", "Validate", :validate)])

  def search
    params = search_params

    @model = JobBreakage.where(job_number: params[:job_number]).order(posted_at: :desc).first;

    return render json: {error: "No breakage found for job #{params[:job_number]}."}, status: :not_found if @model.nil?

    render :show
  end

  def before_set_target
    @model.preload(job_breakage_datum: :job)
  end

  def before_set_targets
    @models.preload(job_breakage_datum: :job)

    # Strip certain filters
    unless params[:filtered].nil?
      params[:filtered].each do |filter|
        if filter[:id] == 'posted_at'
          start_time = resolve_time_filter(filter[:value])
          return @models.where(JobBreakage.arel_table[:posted_at].gteq(start_time))
        end
        if filter[:id] == 'validated'
          return @models.where.not(validated_by: nil) if filter[:value] == 'true'
          return @models.where(validated_by: nil)
        end
      end
    end
    @models
  end

  def before_filter(filter)
    return nil if filter[:id] == 'validated'
    return nil if filter[:id] == 'posted_at'
    filter
  end

  def create
    params = model_params

    @model = JobBreakage.new(
      posted_by: params[:posted_by],
      posted_by_user_id: params[:posted_by_id],
      posted_at: Time.zone.now,
      accepted: params[:accepted],
      validated_by: params[:validated_by],
      validated_at: params[:accepted] ? Time.zone.now : nil,
      job_number: params[:job_number]
    )

    @model.job_breakage_data << build_side_breakage(params[:left]) unless params[:left].nil?
    @model.job_breakage_data << build_side_breakage(params[:right]) unless params[:right].nil?

    if @model.save
      render json: @model, status: :created, location: @model
    else
      render json: @model.errors, status: :unprocessable_entity
    end
  end

  def validate
    params = validate_params

    return render json: {error: "Breakage already validated!"}, status: :unprocessable_entity if @model.is_validated?

    if @model.validate_breakage(params[:accepted], params[:validated_by], params[:validated_by_id])
      render :show
    else
      render json: @model.errors, status: :unprocessable_entity
    end
  end

  private

  def build_side_breakage(data)
    job = Job.find_by_id(data[:job_id])

    raise JobDoesNotExist if job.nil?

    JobBreakageDatum.new(
      job: job,
      broken: data[:broken],
      breakage_reason_id: data[:breakage_reason_id],
      breakage_reason: data[:breakage_reason],
      breakage_category_id: data[:breakage_category_id],
      breakage_category: data[:breakage_category],
    )
  end

  # Only allow a list of trusted parameters through.
  def model_params
    data = params.require(:job_breakage).permit(:posted_by, :posted_by_id, :accepted, :validated_by, :job_number, breakage_data: {})

    {
      posted_by: data[:posted_by],
      posted_by_id: data[:posted_by_id],
      accepted: data[:accepted],
      validated_by: data[:validated_by],
      job_number: data[:job_number],
      right: extract_side_data(data[:breakage_data]['right']),
      left: extract_side_data(data[:breakage_data]['left'])
    }
  end

  def validate_params
    params.permit(:validated_by, :validated_by_id, :accepted)
  end

  def extract_side_data(data)
    return nil if data.nil?

    data.permit(:job_id, :broken, :breakage_reason, :breakage_reason_id, :breakage_category, :breakage_category_id)
  end

  def search_params
    params.permit(:job_number)
  end

  def resolve_time_filter(filter_name)
    return Time.now.at_beginning_of_day if filter_name == 'day'
    return Time.now.at_beginning_of_week if filter_name == 'week'
    return Time.now.at_beginning_of_month if filter_name == 'month'
    return Time.now
  end
end
