class JobsController < ApplicationController
  include MachineResponse

  def self.set_target_before
    super + [:latest_routing_states]
  end

  wrap_parameters :job, include: [:data, :lensside, :processes, :lmatid, :lmattype, :enabled]
  model_data_store(Job)

  def update
    @model = Job.find_by(id: params[:id])

    if @model.nil?
      render status: :unprocessable_entity
    else
      @model.update(model_params)
      render :show
    end
  end

  def routing_history
    ActiveRecord::Base.connected_to(shard: :archive, role: :reading) do
      @limit = params[:limit] || 1000
      @offset = params[:offset] || 0
      @start_date = Time.zone.parse(params[:startTime] || "") || Time.at(0)
      @end_date = Time.zone.parse(params[:endTime] || "")  || Time.current

      @jobs = Job.limit(@limit).offset(@offset)
      render :routing_history
    end
  end

  def add_to_order
    begin
      # Note: The 5 minute delay on the job entry is to ensure orders (separately added) are processed BEFORE their jobs.
      # This is to allow labs that use the order/job/job interface (Optifacts, LensWare) to still use more than one worker process to process the order_entry queue.
      # This time could likely be reduced to 2 minutes but nothing lower. It is critical that the job prio is also LOWER than the order prio.
      AddToOrderJob.set(wait: 5.minute, priority: 1).perform_later(params[:lens_id], add_params, data_params, process_params, params[:job_id])
    rescue => e
      Rails.logger.error e
    end
    send_machine_response
  end

  def delete
    # 28/10/2021 : This endpoint is officially deprecated and only exists for legacy support
    render html: "OK"
  end

  def delete_from_order
    # 28/10/2021 : This endpoint is officially deprecated and only exists for legacy support
    render html: "OK"
  end

  def add
    begin
      AddJob.perform_later(params[:id], add_params, data_params, process_params)
    rescue => e
      Rails.logger.error e
    end
    send_machine_response
  end

  def revive
    @model = Job.find_by(id: params[:job_id])
    if @model.nil?
      head :unprocessable_entity
      return
    end
    if @model.order.present?
      @model.order.revive
    else
      @model.revive
    end
    render :show
  end

  def jobs_status
    ActiveRecord::Base.connected_to(shard: :archive, role: :reading) do
      job_numbers = status_params
      job_status = job_numbers.map do |job_data|
        job_number = job_data[:job_number]
        job = Job.find_by(id: job_data[:rx_data_job_id].to_i)
        if job.nil?
          [job_number, nil]
        else
          [job_number, job.status]
        end
      end.to_h
      render json: job_status
    end
  end

  def states
    @states = Job.find(params[:job_id]).job_states
    @states = @states.where(machine_id: params[:machine_id]) if params[:machine_id]
  end

  def process_data
    job = Job.find(params[:job_id])

    processes = job.job_processes

    if params[:machine_id]
      processes = processes.where(machine_id: params[:machine_id])
    end

    @data = if processes.empty?
      JobProcessParameter.none
    else
      JobProcessParameter.where(job_process_id: processes)
    end
  end

  # Get the jobs tied to the job number scanned on a conveyor tray
  # TODO: This should serve as the entry point for ANY place that needs to get job data using nothing but the tray number (tag data for router)
  def jobs_for_tray_number
    job_number = params[:tray_number]

    # Jobs in a conveyor tray MUST be active
    job = FindHelpers::JobFinders.active_with_number(job_number)

    return render json: {error: "No job with number #{job_number} found."}, status: :not_found if job.nil?

    jobs = {}

    # Use the order to find the lenses tied to the order (jobs in trays MUST be tied to an order)
    job.order.jobs.each do |order_job|
      jobs[order_job.side] = order_job
    end

    render json: {jobs: jobs}
  end

  def job_history
    # If order_id is passed, get jobs belonging to that order directly
    # If only job_number passed, get latest job with that number, get opposite job and return those
    job_number = params[:job_number]
    order_id = params[:order_id]
    if order_id
      order, shard = Order.ensure_find_by_id(order_id)
      jobs_arr = ActiveRecord::Base.connected_to(shard: shard, role: :reading) { order.jobs }
    else
      latest_job, shard = Job.ensure_find_latest_job(job_number)
      return render json: {error: 'noResultsForJob'}, status: :not_found if latest_job.nil?
      job_order = ActiveRecord::Base.connected_to(shard: shard, role: :reading) { latest_job.order }
      return render json: {error: 'jobNotAttachedToOrder'}, status: :not_found if job_order.nil?
      opposite_job = ActiveRecord::Base.connected_to(shard: shard, role: :reading) { latest_job.opposite_job }
      order_id = latest_job.order&.id
      jobs_arr = [latest_job, opposite_job]
    end

    # Get orders both from primary and from archive with that job number
    primary_orders = ActiveRecord::Base.connected_to(shard: :primary, role: :reading) do
      Job.joins(:order).where(number: job_number).select("order_id", "orders.created_at").order("orders.created_at": :desc).distinct(:order_id)
    end

    archive_orders = ActiveRecord::Base.connected_to(shard: :archive, role: :reading) do
      Job.joins(:order).where(number: job_number).select("order_id", "orders.created_at").order("orders.created_at": :desc).distinct(:order_id)
    end

    # Remove duplicate orders from primary and archive
    id_taken = {}
    orders_last_update = []
    (primary_orders + archive_orders).each do |v|
      next if id_taken.has_key? v.order_id
      id_taken[v.order_id] = v
      orders_last_update << v
    end

    # Find which job is left and which job is right
    left_job = jobs_arr.detect {|job| job&.side == :left}
    right_job = jobs_arr.detect {|job| job&.side == :right}

    # Job data from history only needs to follow the same db shard that the order is in.
    # Since if the order is in the archive db, there's no chance that theres data missing from the order in the archive.
    ActiveRecord::Base.connected_to(shard: shard, role: :reading) do
      render json: {
        job_data: {
          left: left_job.nil? ? nil : left_job.data_tags(true),
          right: right_job.nil? ? nil : right_job.data_tags(true)
        },
        machine_history: Job.combine_machine_history(
          left_job.nil? ? nil : left_job.machine_history,
          right_job.nil? ? nil : right_job.machine_history
        ),
        all_orders_update_times: orders_last_update,
        selected_order: order_id
      }, status: :ok
    end
  end

  def latest_routing_states
    # Rx-Routing only keeps the id of one job, so we have to find the other side of the job (if any) through the order
    return render json: {error: 'Job is not attached to any order.'}, status: :unprocessable_entity if @model.order.nil?
    job_states = []
    @model.order.jobs.each do |order_job|
      order_job_states = order_job.job_states.reorder("job_states.time DESC").eager_load(:machine).group_by(&:machine).transform_values(&:first).map do |latest_state|
        {
          machine_sn: latest_state[0].sn,
          side: order_job.side,
          state: latest_state[1].state,
          comment: latest_state[1].comment,
          time: latest_state[1].time
        }
      end
      job_states.concat(order_job_states)
    end
    render json: {job_states: job_states}
  end

  # Keep this for legacy support of Rx-Routing before version 3
  # New version of this method is: /api/orders/routing_launch_data
  def routing_launch_data
    params = routing_launch_params

    job = FindHelpers::JobFinders.latest_attached_to_order_with_number(params[:job_number])

    return render json: {
      error: {
        jobDoesNotExist: true,
        message: 'No job found.'
      }
    }, status: :unprocessable_entity if job.nil?

    return render json: {
      error: {
        jobDoesNotExist: true,
        message: 'Job not attached to order.'
      }
    }, status: :unprocessable_entity if job.order.nil?

    data = job.order.routing_launch_data
    data[:active] = !(job.expired?)
    data[:rx_data_job_id] = job.id

    render json: data, status: :ok
  end

  def breakages
    render json: {breakageList: []}
  end

  private

  def model_params
    params.fetch(:job, {}).permit(:number, :name, :description, :sequence_number, :priority, :expiry_date, :manufacturer, :side, :lmat_id, :lmat_type, :batch_criteria, :template, :job_type)
  end

  def add_params
    job = params.require(:job)
    {
      name: job[:name],
      description: job[:description],
      sequence_number: job[:sequenceNumber],
      priority: job[:priority],
      manufacturer: job[:manufacturer],
      side: Sideable.side_letter_to_side_num(job[:lensside]),
      lmat_id: job[:lmatid],
      lmat_type: job[:lmattype],
      batch_criteria: job[:batchcriteria],
      template: job[:template],
      job_type: job[:jobtype],
      enabled: job[:enabled]
    }
  end

  def data_params
    data = params.require(:job).dig(:data).permit(item: [:name, :value, :lensside]).dig(:item)
    Array(data).map do |datum|
      {
        name: datum[:name],
        value: datum[:value],
        side: Sideable.side_letter_to_side_num(datum[:lensside])
      }
    end
  end

  def process_params
    process_obj = params.require(:job).dig(:processes)
    if process_obj.is_a?(String)
      return []
    end
    processes = process_obj.permit(
      process: [
        :name,
        :enabled,
        plannedQuantity: [:value],
        quantityProductionTime: [:value],
        setupTime: [:value],
        leadTime: [:value],
        capabilities: [
          capability: [:name],
          "capability-group": [:name]
        ]
      ]
    ).dig(:process)
    Array(processes).map do |process|
      # capabilities = process[:capabilities].map do |capability|
      #   cap = Capability.get_by_name(capability[:capability][:name])
      #   cap_group = CapabilityGroup.get_by_name(capability[:"capability-group"][:name])
      #   cap.capability_groups << cap_group
      #   cap
      # end
      {
        name: process[:name],
        planned_quantity: process.dig(:plannedQuantity)&.dig(:value),
        quantity_production_time: process.dig(:quantityProductionTime)&.dig(:value),
        setup_time: process.dig(:setupTime)&.dig(:value),
        lead_time: process.dig(:leadTime)&.dig(:value),
        enabled: process[:enabled]
        # capabilities: capabilities # we're ignoring this until I can find out where it's supposed to go
      }
    end
  end

  def status_params
    params.require(:job_numbers)
  end

  def revive_params
    params.permit(:id)
  end

  def routing_launch_params
    params.permit(:job_number)
  end
end
