import {attribute, DataStore, model, Model} from '@rx-suite/data';
import {t} from "@lingui/macro";
import {action, runInAction} from "mobx";
import {BlockSettings} from "./BlockStore";

@model({
  transportLayer: "/api/block_reuse_settings",
  permissionTarget: 'BlockReuseSetting',
  routingInfo: {
    create: "/block_reuse_settings/create",
    update: "/block_reuse_settings/:id/edit",
    read: "/block_reuse_settings/:id"
  },
  singleton: true
})
export class BlockReuseSettingModel extends Model {

  @attribute({ label: t`Enable ReUse Server`})
  enable_server: boolean = false;

  @attribute({ label: t`Server Port`})
  server_port: number = 0;

  @attribute({ label: t`Wash Block Piece After (Cycles)`})
  block_wash_after: number = 0;

  @attribute({ label: t`Discard Block Piece After (Cycles)`})
  block_discard_after: number = 0;

  @attribute({ label: t`Server Online?`})
  server_online: boolean = false;

  @attribute({ label: t`Block Table Locked?`})
  block_table_locked: boolean = false;

  @attribute({ label: t`Created At`, autoForm: false })
  created_at: Date = new Date();

  @attribute({ label: t`Updated At`, autoForm: false })
  updated_at: Date = new Date();

  @action.bound
  async startReuseServer(): Promise<void> {
    const response = await BlockReuseSettingStore.transportLayer.fetch(`/start_server`, {method: "put"});
    const data = await response.json();
    runInAction(() => {
      const model = this.store.createModel(data);
      this.store.setModel(model);
    })
  }

  @action.bound
  async stopReuseServer(): Promise<void> {
    const response = await BlockReuseSettingStore.transportLayer.fetch(`/stop_server`, {method: "put"});
    const data = await response.json();
    runInAction(() => {
      const model = this.store.createModel(data);
      this.store.setModel(model);
    })
  }

  constructor() {
    super();
    super.setup();
  }
}

type BlockImportData = {
  dmc: string,
  usage_count: number
}

export async function importBlockReUseData(data: string): Promise<void> {
  const blockDataRaw = data.split("\n");

  const blockEntries: BlockImportData[] = [];

  for(let i = 0; i < blockDataRaw.length; i++){
    const blockData = blockDataRaw[i].split(',');
    blockEntries.push({
      dmc: blockData[0],
      usage_count: Number(blockData[1])
    })
  }

  const resp = await fetch("/api/blocks/import", {
    method: "POST",
    headers: {
      "Content-Type": "application/json"
    },
    body: JSON.stringify({ blocks: blockEntries})
  });

  if(!resp.ok){
    throw 'Failed to Import Block ReUse data.'
  }
}

export async function getBlockSettings(): Promise<BlockSettings | null> {
  try {
    const response = await BlockReuseSettingStore.transportLayer.fetch(`/`, {method: 'get'});
    const settings = await response.json();

    const blockSettings = settings.models[0];

    return blockSettings;
  } catch (e) {
    // could not fetch data from rx-data
    return null;
  }
}

export const BlockReuseSettingStore = BlockReuseSettingModel.store as DataStore<BlockReuseSettingModel>;
