import {attribute, DataStore, model, Model} from "@rx-suite/data";

import {t} from "@lingui/macro";
import * as yup from 'yup';
import {action, makeObservable, observable, runInAction} from "mobx";

class MachineStoreDataStore<T extends Model> extends DataStore<T> {
  @observable showArchived = false;

  async toggleArchived(include_archived: boolean) {
    if (this.fetching) {
      return;
    }
    runInAction(() => {
      this.showArchived = include_archived;
    });
    const filter = this.fetchParams.filtered.find(filter => filter.id === 'archived');
    if(!filter){
      this.fetchParams.filtered.push({id: 'archived', value: this.showArchived.toString()});
    }else{
      filter.value = this.showArchived.toString();
    }
    await this.getTableData(true);
  }

  constructor(model, transportLayer, permissionTarget, routingInfo, eager) {
    super(model, transportLayer, permissionTarget, routingInfo, eager);
    makeObservable(this);
  }
}

@model({
  transportLayer: "/api/machines",
  routingInfo: {
    create: "/settings/machines/create",
    update: "/settings/machines/:id/edit",
    read: "/machines/:id/read",
  },
  permissionTarget: 'Machine',
  storeClass: MachineStoreDataStore
})

export class MachineModel extends Model {
  @attribute({label: t`Serial Number`, validation: yup.string().required(t`A Serial Number (SN) must be specified for this machine.`)})
  sn: string = "";

  @attribute({label: t`Name`, validation: yup.string().required(t`A name must be specified for this machine.`).typeError(t`A name must be specified for this machine.`)})
  name: string = "";

  get extendedName(): string {
    return this.name ? `${this.name} (${this.sn})` : `(${this.sn})`;
  }

  @attribute({label: t`Machine Groups`, autoForm: false})
  machine_group_ids = [];

  @attribute({label: t`Machine Type`, validation: yup.number().required().typeError(t`This machine must be assigned to a type.`)})
  machine_type_id: number | null = null;

  @attribute({label: t`Created At`, autoForm: false})
  created_at: Date = new Date();

  @attribute({label: t`Updated At`, autoForm: false})
  updated_at: Date = new Date();

  @attribute({label: t`Use for Conveyor Routing?`, autoForm: false})
  is_conveyor_machine: boolean = true;

  @attribute({label: t`Record EER metrics?`, autoForm: false})
  track_eer_metrics: boolean = false;

  @attribute({label: t`Archived?`, autoForm: false})
  archived: boolean = false;

  @attribute({label: t`Handles Rejects for Machines`, autoForm: false})
  machine_ids: string[] = [];

  @attribute({label: t`Send Push Targets?`, autoForm: false})
  push_target_enabled: boolean = true;

  async jobsForPeriod(period: string) {
    return await getJobsInfo(this.id, this.sn, period);
  }

  async processInfoForJobs(jobId: string) {
    return await getProcessData(this.id, jobId)
  }

  async exportCSV(endpoint: string, startDate: string, endDate: string) {
    return await getCSV(this.id, startDate, endDate, endpoint);
  }

  @action.bound
  async archive(confirmationMessage?: string) {
    let confirmation = true;
    if (confirmationMessage) {
      confirmation = window.confirm(confirmationMessage);
    }
    if (confirmation) {
      this.deletable = true;
      this._state = 'updating';
      const resp = await fetch(`${MachineStore.transportLayer.base}/${this.id}/archive`, {
        method: "PUT",
        headers: {'Content-Type': 'application/json'}
      })
      const data = await resp.json();
      if (!resp.ok) {
        throw data.error;
      }
      this.store.deleteModel(this.getId().toString());
      this.store.totalRecords -= 1;
      if (this.store.fetchParams.limit !== undefined && this.store.tableData.length < this.store.fetchParams.limit) {
        this.store.getTableData(true);
      }
    }
  }

  @action.bound
  async restore(confirmationMessage?: string) {
    let confirmation = true;
    if (confirmationMessage) {
      confirmation = window.confirm(confirmationMessage);
    }
    if (confirmation) {
      this._state = 'updating';
      const resp = await fetch(`${MachineStore.transportLayer.base}/${this.id}/restore`, {
        method: "PUT",
        headers: {'Content-Type': 'application/json'}
      });
      const data = await resp.json();
      if (!resp.ok) {
        throw data.error;
      }
      this.store.getTableData(true);
      this._state = 'done';
    }
  }

  constructor() {
    super();
    super.setup();
  }
}

async function getJobsInfo(machineId: string, machineSn: string, period: string) {
  const periodAmount = (function (period) {
    switch(period) {
      case 'day':
        return 1;
      case 'week':
        return 7;
      case 'month':
        return 30;
    }
  })(period);

  let d = new Date();
  d.setDate(d.getDate() - periodAmount);
  const endTime = encodeURIComponent(new Date().toISOString().replace('T', ' ').replace('Z',''));
  const startTime = encodeURIComponent(d.toISOString().replace('T', ' ').replace('Z',''));
  const params = `startTime=${startTime}&endTime=${endTime}`;
  const connectorUrl = '/connector//satisloh/reportqueries/resource';

  const jobResp = await fetch(`${connectorUrl}/${machineSn}/1.0/jobs?${params}`);
  const job = await jobResp.json();

  return job.processList;
}

async function getProcessData(machineId: string, jobId: string) {
  const processDataResp = await fetch(`/api/jobs/${jobId}/process_data?machine_id=${machineId}`);
  const statesResp = await fetch(`/api/jobs/${jobId}/states?machine_id=${machineId}`);
    const processData = await processDataResp.json();
    const states = await statesResp.json();
    return {
      processDataList: processData,
      statusList: states
    }
}

async function getCSV(machineId: string, startDate: string, endDate: string, endpoint: string) {

  const resp = await fetch(`/api/export/machines/${machineId}/${endpoint}`, {
    method: "POST",
    headers: {
      "Content-Type": "application/json"
    },
    body: JSON.stringify({ start_date: startDate, end_date: endDate })
  });

  const csv = await resp.json();

  return csv;
}

export async function getProductionData(machineId: string) {
  const resp = await fetch(`/api/machines/${machineId}/production_data`);
  return await resp.json();
}

export async function getLastMachineState(machineId: string) {
  const resp = await fetch(`/api/machines/${machineId}/machine_last_state`);
  return await resp.json();
}

export const MachineStore = MachineModel.store as MachineStoreDataStore<MachineModel>;

