class AddBoxCoaterEventJob < ApplicationJob
  queue_as :eer

  def perform(machine_state)
    @machine_state = machine_state
    @event_type = machine_state.box_coater_event_type

    machine_scope = BoxCoaterEvent.where(machine: @machine_state.machine)
    @prior_event = machine_scope.where(BoxCoaterEvent.arel_table[:time].lteq(@machine_state.time)).order(time: :desc).limit(1).first
    @next_event = machine_scope.where(BoxCoaterEvent.arel_table[:time].gt(@machine_state.time)).order(time: :asc).limit(1).first

    if is_duplicate_time?(machine_state)
      return Rails.logger.error "Received box coater door state with exact time as previous state! Ignoring state change."
    end

    # During engineering lockouts, we ignore any door transitions
    unless @prior_event.nil?
      return if @prior_event.is_engineering_lockout?
    end

    # If there's no prior EER event or this event type is different than the next, create a new EER event
    if @prior_event.nil? || @prior_event.event_type != @event_type
      create_new_event
    else
      update_ongoing_event
    end
  end

  def create_new_event
    # Close the previous event
    @prior_event.update(duration: @machine_state.time - @prior_event.time) unless @prior_event.nil?

    # Latest event
    if @next_event.nil?
      new_event = BoxCoaterEvent.new(machine: @machine_state.machine, event_type: @event_type, recipe: @machine_state.recipe, time: @machine_state.time)

      batch = new_event.find_or_create_associated_batch
      unless batch.nil? # Door Open events have no batches to assign
        new_event.box_coater_batch = batch
        batch.box_coater_event = new_event
      end

      if new_event.save
        new_event.broadcast
      end
    # Event came out of order, save but don't do anything
    else
      BoxCoaterEvent.create(machine: @machine_state.machine, event_type: @event_type, recipe: @machine_state.recipe, time: @machine_state.time, duration: @next_event.time - @machine_state.time)
    end
  end

  def is_duplicate_time?(machine_state)
    return false if @prior_event.nil?

    # If somehow the prior event has the EXACT time as the new event, ignore this new one, we have a problem.
    return @prior_event.time == machine_state[:time]
  end

  def update_ongoing_event
    return unless @prior_event.is_door_closed?
    return unless @prior_event.recipe.nil?
    return unless @machine_state.has_recipe?

    @prior_event.assign_attributes({recipe: @machine_state.recipe})

    if @prior_event.save
      @prior_event.broadcast if @next_event.nil?
    end
  end
end
