module BucketCalculator
  include KpiHelpers

  class InvalidTimeFrame < StandardError; end

  # Returns a hash of bucket with the start time being the key of each bucket
  # Example: buckets for 24 hrs
  # buckets = {
  # '01/01/2022 - 13h00' : 0,
  # '01/01/2022 - 13h30' : 0,
  # '01/01/2022 - 14h00' : 0
  # ...
  # }
  def time_buckets(start_time, end_time, timeframe, bucket_default_val: 0)
    # If we have timeframe all, then there's no time buckets, just return the default values
    return bucket_default_val if timeframe_all?(timeframe)

    buckets = get_buckets(start_time, end_time, timeframe).map { |b| b[:start_time]}
    Hash[buckets.zip(Array.new(buckets.size) {bucket_default_val.dup})]
  end

  # Returns an array of bucket with the start time of each bucket
  # Example: buckets for 24 hrs
  # buckets = [
  #   {start: '01/01/2022 - 13h00'},
  #   {start: '01/01/2022 - 13h30'},
  #   {start: '01/01/2022 - 14h00'}
  #   ...
  # ]
  def get_cache_bucket_times(start_time, end_time)
    cache_bucket_times = []
    i_time = KpiHelpers.kpi_bucket_time(start_time)
    e_time = KpiHelpers.kpi_bucket_time(end_time)
    # end_time should be with 59 minutes and 59 seconds format, so e_time should be included
    while i_time <= e_time
      cache_bucket_times.push(i_time)
      i_time += KpiHelpers.cache_entry_size
    end
    cache_bucket_times
  end

  def get_buckets(start_time, end_time, timeframe)
    num_buckets, time_between_buckets = bucket_calc(timeframe, start_time, end_time)
    (0..num_buckets-1).map do |i|
      {
        start_time: start_time + (i * time_between_buckets),
        end_time: start_time + ((1+i) * time_between_buckets) - 1
      }
    end
  end

  def bucket_calc(time_frame, start_time, end_time)
    case time_frame
    when Timeframe::HOUR_BY_HOUR
      return [hours_between(start_time, end_time), 1.hour]
    when Timeframe::DAY_BY_DAY
      # we go off of dates here so that if we have our start_time at 8 pm and end_time at 8 am the next day it would
      # just count as one bucket but it spans two days, so 2 buckets is necessary
      return [(start_time.to_date..end_time.to_date).count, 1.day]
    when Timeframe::WEEK_BY_WEEK
      return [weeks_between(start_time, end_time), 1.week]
    when Timeframe::ALL
      return [1, end_time - start_time]
    else
      raise InvalidTimeFrame
    end
  end

  def hours_between(start_time, end_time)
    ((end_time-start_time) / 3600.0).ceil
  end

  def days_between(start_time, end_time)
    (hours_between(start_time, end_time) / 24.0).ceil
  end

  def weeks_between(start_time, end_time)
    (days_between(start_time, end_time) / 7.0).ceil
    #https://stackoverflow.com/a/7639684
    #(0.5 + (start_time.end_of_month.day + start_time.at_beginning_of_month.wday).to_f / 7).ceil
  end
end
