module KpiTableHandler
  module Base
    # KPI cache deletions have to be managed so that the 'current' record isn't deleted while a thread is using
    # it, otherwise the result of the query would be inaccurate (for example missing entries). This class manages
    # the deletion per table by not allowing two deletes to happen in parallel and keeps track of the last time
    # the rows for a machine were deleted. If data is fresh for the request row, the delete will not be triggered
    class KpiCacheManager
      include KpiHelpers

      class << self
        def delete_dirty_rows(kpi_table, start_time, end_time)
          # If the cache cannot be dirty, don't bother going further
          return false unless kpi_table.column_names.include? 'dirty'

          synchronize(kpi_table) do
            return unless delete_required?(kpi_table)

            kpi_table.where(dirty: true)
                     .where(kpi_table.arel_table[:time].gteq(start_time).and(kpi_table.arel_table[:time].lt(end_time)))
                     .delete_all

            refresh_delete_time(kpi_table)
          end
        end

        def machine_kpi_row_last_refresh_time(kpi_table)
          manager = table_manager(kpi_table)
          manager[:last_delete_cycle]
        end

        private

        def table_manager(kpi_table)
          @_kpi_table_manager ||= {}
          @_kpi_table_manager[kpi_table] ||= {
            mutex: Mutex.new,
            last_delete_cycle: Time.zone.now - 1.hour
          }
        end

        def refresh_delete_time(table)
          manager = table_manager(table)
          manager[:last_delete_cycle] = Time.zone.now
        end

        def delete_required?(table)
          manager = table_manager(table)

          # If the current time belongs to a different bucket than the last time the cache manager was updated, we need to perform an update
          return true if KpiHelpers.kpi_bucket_time(Time.zone.now) != KpiHelpers.kpi_bucket_time(manager[:last_delete_cycle])

          Time.zone.now - manager[:last_delete_cycle] > cache_stale_time
        end

        def synchronize(kpi_table, &block)
          table_manager(kpi_table)[:mutex].synchronize do
            block.call
          end
        end

        def cache_stale_time
          2.minutes
        end
      end
    end
  end
end