module MachineStateDurationHelpers
  extend ActiveSupport::Concern

  include KpiHelpers
  include KpiTableHandler
  include ArchiveHelpers

  def build_machine_state_duration_kpi_buckets(start_time, end_time, leave_empty=false)
    buckets = empty_bucket_for_interval(start_time, end_time)

    return format_buckets_for_cache(buckets) if leave_empty

    ensure_machine_state_durations(start_time, end_time)

    # Handle state that starts before the interval
    state_at_start = state_prior_to_start(start_time)

    if state_at_start.nil?
      adjust_first_bucket(buckets, start_time, end_time)
    else
      apply_state_duration_to_buckets(buckets, state_at_start, start_time)

      # Since there's a single ongoing state, we can exit now
      return format_buckets_for_cache(buckets) if state_at_start.next_in_time.nil?
    end

    # Machine states entirely within the interval
    MachineState.where(machine_sn: sn)
                .where(MachineState.arel_table[:time].gt(start_time).and(MachineState.arel_table[:time].lt(end_time)))
                .order(time: :asc)
                .find_each do |machine_state|
      apply_state_duration_to_buckets(buckets, machine_state, start_time)
    end

    format_buckets_for_cache(buckets)
  end

  private

  def empty_bucket_for_interval(start_time, end_time)
    buckets = {}
    KpiHelpers.time_buckets_between(start_time, end_time).each do |time|
      buckets[time] = new_machine_state_cache_entry(time)
    end
    buckets
  end

  def new_machine_state_cache_entry(start_time)
    available_time = if KpiHelpers.bucket_is_current_time?(start_time)
                       KpiHelpers.machine_state_duration_current_bucket_size
                     elsif start_time > Time.zone.now
                       0
                     else
                       KpiHelpers.cache_entry_size_seconds
                     end

    {
      available_time: available_time,
      machine_id: id,
      time: start_time,
      dirty: KpiHelpers.bucket_is_current_time?(start_time),
      on: 0,
      notready: 0,
      ready: 0,
      waitingforjobs: 0,
      processing: 0,
      maintenance: 0,
      error: 0,
      halt: 0,
      estop: 0,
      setup: 0,
      off: 0
    }
  end

  # Ensure each machine state has a calculated duration before moving forward with the rest of the calc.
  # note: since old machine states might not have calculated durations, we need to write to the archive db directly
  def ensure_machine_state_durations(start_time, end_time)
    # Removing a day since @interval for build archive is archive_time + 1.day (we were trying to write to read only connection to archive)
    if time_exists_in_primary?(start_time - 1.days)
      build_machine_state_duration(start_time, end_time)
    else
      ActiveRecord::Base.connected_to(shard: :archive, role: :writing) do
        build_machine_state_duration(start_time, end_time)
      end
    end
  end

  def build_machine_state_duration(start_time, end_time)
    MachineState.where(machine_sn: sn, calculated_duration: nil)
                .where(MachineState.arel_table[:time].gteq(start_time).and(MachineState.arel_table[:time].lteq(end_time)))
                .find_each do |machine_state|
      next_state = machine_state.next_in_time
      # since this can cause several models to updated, skip validation and callbacks by invoking update_columns directly
      machine_state.update_columns(calculated_duration: next_state.time - machine_state.time) unless next_state.nil?
    end
  end

  # Find the state that starts prior to the interval, this state can be in the archive even if we are searching for an interval in the primary
  def state_prior_to_start(start_time)
    state_at_start = MachineState.where(machine_sn: sn).where(MachineState.arel_table[:time].lteq(start_time)).order(time: :desc).limit(1).first

    if state_at_start.nil?
      ActiveRecord::Base.connected_to(shard: :archive, role: :reading) do
        state_at_start = MachineState.where(machine_sn: sn).where(MachineState.arel_table[:time].lteq(start_time)).order(time: :desc).limit(1).first
      end
    end

    state_at_start
  end

  # If there is no state prior to the start of the requested interval, the first bucket has 'empty' duration that must be removed.
  # ex: the first machine state ever is at 13h07. The time from 13h00 to 13h07 must be removed.
  def adjust_first_bucket(buckets, start_time, end_time)
    # happens in the case where start_time is in the future, available time is already considered to be zero
    return if buckets.first[1][:available_time] == 0

    first_state = MachineState.where(machine_sn: sn)
                              .where(MachineState.arel_table[:time].gt(start_time).and(MachineState.arel_table[:time].lt(end_time)))
                              .order(time: :asc).limit(1).first

    return if first_state.nil?

    buckets.first[1][:available_time] = KpiHelpers.cache_entry_size - (first_state.time - buckets.first[0])
  end

  # While doing an insert all, we cannot have invalid columns, so remove the extra available time col.
  def format_buckets_for_cache(buckets)
    buckets.map { |k, v| v.except!(:available_time) }
  end

  def apply_state_duration_to_buckets(buckets, machine_state, start_time)
    duration = machine_state_relative_duration(machine_state, start_time)
    bucket_time = machine_state_start_bucket(machine_state, start_time)

    while duration > 0
      bucket = buckets[bucket_time]
      break if bucket.nil?

      applicable_time = [duration, bucket[:available_time]].min
      bucket[machine_state.state.to_sym] += applicable_time
      bucket[:available_time] -= applicable_time
      duration -= applicable_time
      bucket_time += KpiHelpers.cache_entry_size
    end
  end

  # Adjust the machine state duration to match the start interval and calculated ongoing machine states
  def machine_state_relative_duration(machine_state, start_time)
    if machine_state.time < start_time
      return machine_state.next_in_time.time - start_time if machine_state.next_in_time

      Time.zone.now - start_time
    else
      return machine_state.calculated_duration unless machine_state.calculated_duration.nil?

      Time.zone.now - machine_state.time
    end
  end

  def machine_state_start_bucket(machine_state, start_time)
    return KpiHelpers.kpi_bucket_time(machine_state.time) if machine_state.time >= start_time

    KpiHelpers.kpi_bucket_time(start_time)
  end
end
