module ReportComponents
    class MachineStateBreakdownPie < ReportComponent
      include DataHelpers
      include KpiHelpers

      attr_accessor :states, :colors

      def component_id
        'machine_state_breakdown'
      end

      def component_type
        'pie'
      end

      def generate(args)
        @states = MachineKpi.state_breakdown_between(start_time,
                                                     end_time,
                                                     Timeframe::ALL,
                                                     machine_ids,
                                                     options: args.merge({ addons: [:as_percent] }))

        # Preload all the colors as we'll need them all regardless
        @colors = Color.all.pluck(:status_name, :color)

        series, labels, colors = if ProductionSetting.use_simple_machine_state_colors?
                                   build_simplified_graph_data
                                 else
                                   build_graph_data
                                 end

        series, labels, colors = ensure_data(series, labels, colors)

        return set_no_data('No machine states during reporting period') if series.blank?

        {
          series: series,
          labels: labels,
          colors: colors
        }
      end

      def ensure_data(series, labels, colors)
        return [[], [], []] if series.blank? && labels.blank? && colors.blank?
        series, labels, colors = series.zip(labels, colors).sort.reverse!.transpose
        data_sum = series.sum

        # avoid division by zero
        # don't pad a margin of error
        not_connected_percentage = 100 - data_sum
        not_connected_time = if use_planned_production_time
                               ShiftCache.get_active_time_between(start_time, end_time) * not_connected_percentage/100
                             else
                               (end_time - start_time) * not_connected_percentage/100
                             end

        if data_sum > 0 && (not_connected_percentage > 1)
          series << (not_connected_percentage.round(2))
          labels << "notconnected - #{time_annotation(not_connected_time)}"
          colors << '#ABB2B9'
        end
        ## NOTE: At the moment, if there is unresolved data, it is not displayed, instead, the existing data is normalized.
        ## series = series.map { |val| ((val * 100) / data_sum).round(2)} if (data_sum > 0 && (100 - data_sum > 1))
        [series, labels, colors]
      end

      def build_graph_data
        series = []
        labels = []
        colors = []

        @states.each do |state, value|
          next unless value[:percent] > 0

          series << value[:percent]
          labels << state.to_s + " - #{time_annotation(value[:count])}"
          colors << color_for_state(state)
        end

        [series, labels, colors]
      end

      def build_simplified_graph_data
        series = []
        labels = []
        colors = []

        groups = [
          [:processing],
          [:waitingforjobs],
          [:estop, :halt, :notready, :error],
          [:calibration, :maintenance, :on, :setup, :ready],
          [:off]
        ]

        groups.each do |group|
          value = value_for_machine_state_group(group)
          next unless value[:percent] > 0

          series << value[:percent]
          labels << group.join(" - ") + " - #{time_annotation(value[:count])}"
          colors << color_for_state(group[0])
        end

        [series, labels, colors]
      end


      # Only include label if at least one machine state is present and it's value is greater than 0
      def value_for_machine_state_group(states)
        acc = {
          count: 0,
          percent: 0
        }
        states.each do |state|
          next if @states[state].nil? || @states[state][:percent].nil? || @states[state][:count].nil?
          acc.merge!(@states[state]) { |k, a, b| a + b}
        end
        acc
      end

      def color_for_state(state)
        color = @colors.find{|c| c[0].downcase == state.to_s.downcase }
        if color.nil?
          '#ABB2B9'
        else
          color[1]
        end
      end

      def time_annotation(value)
        prefix = number_of_machines > 1 ?
                   'avg ' :
                   ''

        suffix = {
          Timeframe::HOUR_BY_HOUR => lambda {
            time = value/60/number_of_machines
            return "< 1 min" if time < 1
            "#{(time).round} mins"
          },
          Timeframe::DAY_BY_DAY => lambda {
            time = value/3600/number_of_machines
            return "< 1 hour" if time < 1
            "#{(time).round(1)} hours"
          },
          Timeframe::WEEK_BY_WEEK => lambda {
            time = value/86400/number_of_machines
            return "< 1 day" if time < 1
            "#{(time).round(1)} days"
          }
        }
        prefix + suffix[timeframe].call
      end
    end
end

